from __future__ import annotations

import textwrap
import typing as t


class CodeDef:
    __slots__ = ("code", "message", "example", "hidden", "default_disabled")

    def __init__(
        self,
        code: str,
        message: str,
        example: str,
        hidden: bool = False,
        default_disabled: bool = False,
    ) -> None:
        self.code = code
        self.message = message
        self.example = textwrap.dedent(example).strip("\n")
        self.hidden = hidden
        self.default_disabled = default_disabled

    def __str__(self) -> str:
        return f"{self.message} ({self.code})"

    @property
    def category(self) -> str:
        return self.code[0]


ALL_CODES: list[CodeDef] = [
    # internal codes
    CodeDef("X001", "unparsable file", "foo(", hidden=True),
    CodeDef(
        "X002",
        "reached recursion limit during CST checks",
        "# see chardet",
        hidden=True,
    ),
    # errors & warnings, grouped by topic numerically
    # string concat
    CodeDef("E100", "unnecessary string concat", 'x = "foo " "bar"'),
    CodeDef("E101", "unnecessary string concat with plus", 'x = "foo " + "bar"'),
    # ast matching
    CodeDef(
        "W200",
        "two AST branches have identical contents",
        """
        if x is True:
            return y + 1
        else:
            # some comment
            return y + 1
        """,
    ),
    CodeDef(
        "W201",
        "two AST branches have identical trivial contents",
        """
        if x is True:
            return
        else:
            return
        """,
        default_disabled=True,
    ),
    CodeDef(
        "W202",
        "two non-adjacent AST branches have identical contents",
        """
        if x is True:
            return foo(bar())
        elif y is True:
            return 0
        elif z is True:
            return 1
        else:
            return foo(bar())
        """,
        default_disabled=True,
    ),
    CodeDef(
        "W203",
        "two non-adjacent AST branches have identical trivial contents",
        """
        if x is True:
            return None
        elif y is True:
            return 0
        elif z is True:
            return 1
        else:
            return None
        """,
        default_disabled=True,
    ),
]
CODE_MAP: dict[str, CodeDef] = {c.code: c for c in ALL_CODES}


def _generate_reference_raw() -> t.Iterator[str]:
    yield (
        """\
E is for "error" (you should probably change this)

W is for "warning" (you might want to change this)

Some warnings are disabled by default; enable them with ``--enable``.
"""
    )

    for code in ALL_CODES:
        if code.hidden:
            continue
        yield ""
        yield f"{code.code}"
        yield f"{'-' * len(code.code)}"
        yield ""
        if code.default_disabled:
            yield "*disabled by default*"
            yield ""
        yield f"{code.message}"
        yield ""
        yield ".. code-block:: python"
        yield ""
        yield textwrap.indent(code.example, "    ")


def generate_reference() -> t.Iterator[str]:
    for s in _generate_reference_raw():
        if s == "":
            yield "\n"
        else:
            for sub in s.splitlines():
                yield sub + "\n"



from slyp.codes import generate_reference

START_SLUG = ".. generate-reference-insert-start\n"
END_SLUG = ".. generate-reference-insert-end\n"


def main() -> None:
    print("update-generated-reference:BEGIN")
    if update_linter_rules():
        print("changes made")
    else:
        print("no changes made")
    print("update-generated-reference:END")


def update_linter_rules() -> bool:
    with open("docs/linter_rules.rst", encoding="utf-8") as fp:
        content = fp.readlines()
    new_content = []

    for line in content:
        new_content.append(line)
        if line == START_SLUG:
            break
    new_content.append("\n")
    new_content.extend(generate_reference())
    found_end_line = False
    for line in content:
        if found_end_line:
            new_content.append(line)
        elif line == END_SLUG:
            found_end_line = True
            new_content.extend(("\n", line))
    if new_content == content:
        return False

    with open("docs/linter_rules.rst", "w", encoding="utf-8") as fp:
        fp.write("".join(new_content))
    return True


if __name__ == "__main__":
    main()



import datetime
import importlib.metadata

project = "slyp"
copyright = f"2023-{datetime.datetime.today().strftime('%Y')}, Stephen Rosen"
author = "Stephen Rosen"

# The full version, including alpha/beta/rc tags
release = importlib.metadata.version("slyp")

extensions = ["sphinx_issues"]
issues_github_path = "sirosen/slyp"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build"]

# HTML theme options
html_theme = "furo"
pygments_style = "friendly"
pygments_dark_style = "monokai"  # this is a furo-specific option
html_theme_options = {
    "source_repository": "https://github.com/sirosen/slyp/",
    "source_branch": "main",
    "source_directory": "docs/",
}



import textwrap

import pytest

from slyp.checkers import _clear_errors, check_file
from slyp.fixer import fix_file
from slyp.hashable_file import HashableFile


@pytest.fixture
def check_text(tmpdir):
    _clear_errors()

    def _check_text(
        text,
        *,
        disabled_codes=None,
        enabled_codes=None,
        dedent=True,
        filename="X.py",
    ):
        handle = tmpdir.join(filename)
        text = textwrap.dedent(text) if dedent else text
        handle.write(text)

        with tmpdir.as_cwd():
            return check_file(
                HashableFile(filename),
                disabled_codes=disabled_codes or set(),
                enabled_codes=enabled_codes or set(),
            )

    return _check_text


@pytest.fixture
def fix_text(tmpdir):
    def _fix_text(text, *, expect_changes=True, dedent=True, filename="X.py"):
        handle = tmpdir.join(filename)
        text = textwrap.dedent(text) if dedent else text
        handle.write(text)

        with tmpdir.as_cwd():
            res = fix_file(HashableFile(filename))
            new_text = handle.read()

            if expect_changes:
                assert not res.success, "fixing made no changes"
            else:
                assert res.success, "fixing made changes"
                assert text == new_text

            return new_text, res

    return _fix_text



from __future__ import annotations

import argparse
import glob
import hashlib
import json
import multiprocessing.pool
import os
import stat
import subprocess
import sys
import time
import typing as t

from slyp.checkers import check_file
from slyp.codes import CODE_MAP
from slyp.file_cache import PassingFileCache
from slyp.fixer import fix_file
from slyp.hashable_file import HashableFile
from slyp.result import Message, Result

DEFAULT_DISABLED_CODES: set[str] = {"W201", "W202", "W203"}
CONTRACT_VERSION: str = "1.6"


def driver_main(args: argparse.Namespace) -> bool:
    # parse inputs from comma delimited lists
    disabled_codes = {x for x in args.disable.split(",") if x != ""}
    enabled_codes = {x for x in args.enable.split(",") if x != ""}
    # add default disables if "all" is not in --enable
    if "all" not in enabled_codes:
        disabled_codes = disabled_codes | DEFAULT_DISABLED_CODES

    if args.files == ["-"]:
        return process_stdin(args, disabled_codes, enabled_codes)
    else:
        return parallel_process(args, disabled_codes, enabled_codes)


def process_stdin(
    args: argparse.Namespace, disabled_codes: set[str], enabled_codes: set[str]
) -> bool:
    result = Result(success=True, messages=[])
    file_obj = HashableFile("-")

    if args.only == "fix":
        result = fix_file(file_obj)
        message_stream = sys.stderr
    elif args.only == "lint":
        result = result.join(
            check_file(
                file_obj,
                disabled_codes=disabled_codes,
                enabled_codes=enabled_codes,
            )
        )
        message_stream = sys.stdout
    else:
        raise NotImplementedError("stdin with unexpected --only value")

    for message in result.messages:
        if message.verbosity <= args.verbosity:
            print(message.message, file=message_stream)

    return result.success


def parallel_process(
    args: argparse.Namespace, disabled_codes: set[str], enabled_codes: set[str]
) -> bool:
    if not args.no_cache:
        passing_cache: PassingFileCache | None = PassingFileCache(
            contract_version=CONTRACT_VERSION,
            config_id=compute_config_id(enabled_codes, disabled_codes),
        )
    else:
        passing_cache = None

    process_pool = multiprocessing.pool.Pool()

    futures = {}
    for filename in all_py_filenames(args.files, args.use_git_ls):
        if args.verbosity >= 1:
            print(f"slpy: processing {filename}", file=sys.stderr)
        futures[filename] = process_pool.apply_async(
            process_file,
            (filename, args.only, disabled_codes, enabled_codes, passing_cache),
        )
    process_pool.close()

    success = True

    while futures:
        ready = set()
        for filename, future in futures.items():
            if future.ready():
                ready.add(filename)
        if not ready:
            time.sleep(0.05)

        for filename in ready:
            future = futures.pop(filename)
            try:
                result = future.get()
            except Exception as e:
                result = Result(
                    success=False,
                    messages=[
                        Message(f"slyp error on '{filename}': {e}"),
                        Message(
                            f"slyp error on '{filename}': {e.__traceback__}",
                            verbosity=2,
                        ),
                    ],
                )
            for message in result.messages:
                if message.verbosity <= args.verbosity:
                    print(message.message)

            success = success and result.success

    process_pool.join()

    return result.success


def process_file(
    filename: str,
    only: str | None,
    disabled_codes: set[str],
    enabled_codes: set[str],
    passing_cache: PassingFileCache | None,
) -> Result:
    result = Result(success=True, messages=[])
    file_obj = HashableFile(filename)

    if passing_cache:
        if file_obj in passing_cache:
            result.messages.append(
                Message(message=f"cache hit: {filename}", verbosity=2)
            )
            return result

    if only in ("fix", None):
        result = result.join(fix_file(file_obj))
    if only in ("lint", None):
        result = result.join(
            check_file(
                file_obj, disabled_codes=disabled_codes, enabled_codes=enabled_codes
            )
        )

    if passing_cache and result.success and only is None:
        passing_cache.add(file_obj)
    return result


def compute_config_id(enabled_codes: set[str], disabled_codes: set[str]) -> str:
    # now we get the codes which are defined, convert to a string
    all_codes: str = json.dumps(sorted(CODE_MAP.keys()))
    # get the enabled/disabled codes, and make that a string
    code_opts = json.dumps(
        {
            "disabled": sorted(enabled_codes),
            "enabled": sorted(disabled_codes),
        }
    )

    config_hash = hashlib.sha256()
    config_hash.update(all_codes.encode())
    config_hash.update(code_opts.encode())

    # full ID is the base + the computed bits hashed
    return config_hash.hexdigest()


def all_py_filenames(files: t.Sequence[str], use_git_ls: bool) -> t.Iterable[str]:
    if files:
        yield from files
    elif use_git_ls:
        git_ls_files_proc = subprocess.run(
            ["git", "ls-files"], check=True, capture_output=True, text=True
        )
        candidates = git_ls_files_proc.stdout.split("\n")
        for file in candidates:
            if file == "":
                continue
            if is_python(file):
                yield file
    else:
        yield from glob.glob("**/*.py", recursive=True)


def is_python(filename: str) -> bool:
    # don't pay attention to symlinks or other special filetypes
    # those aren't really files to check
    # but you might be fed them by `git ls-files`
    if not stat.S_ISREG(os.lstat(filename).st_mode):
        return False

    # .py is good
    if filename.endswith(".py"):
        return True

    # if it's not executable, it couldn't be a script
    if not os.access(filename, os.X_OK):
        return False

    # but otherwise, look for a shebang which contains 'python' as a substring
    try:
        with open(filename, encoding="utf-8") as fp:
            firstline = fp.readline()
    except UnicodeDecodeError:
        return False
    if firstline.startswith("#!") and "python" in firstline:
        return True

    return False



import textwrap


def test_unannotated_init_gets_return_none_added(fix_text):
    new_text, _ = fix_text(
        """\
        class A:
            def __init__(self): ...
        """
    )
    assert new_text == textwrap.dedent(
        """\
        class A:
            def __init__(self) -> None: ...
        """
    )


def test_annotated_init_sees_no_change(fix_text):
    fix_text(
        """\
        class A:
            def __init__(self) -> None: ...
        """,
        expect_changes=False,
    )


def test_badly_annotated_init_sees_no_change(fix_text):
    # this annotation is fundamentally *wrong*, but it should be ignored by the fixer
    fix_text(
        """\
        class A:
            def __init__(self) -> A: ...
        """,
        expect_changes=False,
    )


def test_mixed_arg_annotations_dont_change_init_type(fix_text):
    new_text, _ = fix_text(
        """\
        class A:
            def __init__(self, x: int, y): ...
        """
    )
    assert new_text == textwrap.dedent(
        """\
        class A:
            def __init__(self, x: int, y) -> None: ...
        """
    )



import libcst


class ErrorCollectingVisitor(libcst.CSTVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.filename: str = "<unset>"
        self.errors: set[tuple[int, str, str]] = set()



import textwrap

import pytest


@pytest.mark.parametrize(
    "original_text, fixed_text",
    [
        ("dict()", "{}"),
        ("list()", "[]"),
        ("tuple()", "()"),
    ],
)
def test_builtin_call_fixer_converts_empty_call_to_empty_literal(
    fix_text, original_text, fixed_text
):
    new_text, _ = fix_text(f"x = {original_text}")
    assert new_text == f"x = {fixed_text}"


def test_dict_call_fixer_converts_kwargs_to_quoted_keys(fix_text):
    new_text, _ = fix_text(
        """\
        a = dict(x=1, y="foo", z_arg={})
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = {"x": 1, "y": "foo", "z_arg": {}}
        """
    )


def test_dict_call_fixer_preserves_double_star_expansion(fix_text):
    new_text, _ = fix_text(
        """\
        a = dict(x=1, **kwargs)
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = {"x": 1, **kwargs}
        """
    )


def test_dict_call_fixer_handles_multi_doublestar(fix_text):
    new_text, _ = fix_text(
        """\
        a = dict(x=1, **kwargs, **kwargs2)
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = {"x": 1, **kwargs, **kwargs2}
        """
    )


def test_dict_call_fixer_does_not_munge_positional_arg(fix_text):
    fix_text(
        """\
        a = dict([("x", 1)])
        b = dict([("x", 1)], y=2)
        c = dict(some_func(), y=2)
        d = dict(some_func(), **other)
        e = dict(some_name, **other)
        f = dict(some_name)
        """,
        expect_changes=False,
    )


def test_dict_call_fixer_handles_nested_calls(fix_text):
    new_text, _ = fix_text(
        """\
        a = dict(x=dict(y=2))
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = {"x": {"y": 2}}
        """
    )


def test_dict_call_fixer_preserves_the_simplest_whitespace(fix_text):
    # trailing comma whitespace + leading whitespace is the simple case
    # because the trailing comma carries the whitespace info, it's easy to
    # preserve by just keeping the comma intact
    new_text, _ = fix_text(
        """\
        a = dict(
            x=dict(
                y=2,
            ),
            z=dict(
                **p,
            )
        )
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = {
            "x": {
                "y": 2,
            },
            "z": {
                **p,
            }
        }
        """
    )


def test_dict_call_fixer_preserves_intricate_whitespace(fix_text):
    # no trailing comma, or empty block are more intricate cases
    # we need to find the whitespace which is attached to the last arg
    # and reattach it to the rbrace
    new_text, _ = fix_text(
        """\
        a = dict(
            x=dict(
                    y="hi"
            ),
            z=dict(

            ),
            p=dict(

                **q
        )
        )
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = {
            "x": {
                    "y": "hi"
            },
            "z": {

            },
            "p": {

                **q
        }
        }
        """
    )


def test_set_fixer_converts_call_of_generator(fix_text):
    new_text, _ = fix_text(
        """\
        a = set(x for x in y())
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = {x for x in y()}
        """
    )


def test_list_fixer_converts_call_of_generator(fix_text):
    new_text, _ = fix_text(
        """\
        a = list(x for x in y())
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = [x for x in y()]
        """
    )


def test_generator_comp_fixer_requires_exactly_one_arg(fix_text):
    # these usages are syntactically valid but semantically incorrect
    # however, the fixer should ignore them as "not my problem"
    fix_text(
        """\
        a = set((x for x in y()), True)
        b = list((x for x in y()), True)
        """,
        expect_changes=False,
    )


@pytest.mark.parametrize("set_variant", ("set", "frozenset"))
@pytest.mark.parametrize(
    "expression, unwrapped",
    (
        ("set(foo)", "foo"),
        ("list(foo)", "foo"),
        ("sorted(foo)", "foo"),
        ("sorted(foo, reversed=True)", "foo"),
        ("reversed(foo)", "foo"),
        ("tuple(foo)", "foo"),
        ("set(tuple(foo))", "foo"),
        ("set()", ""),
        ("list()", ""),
        ("tuple()", ""),
        ("[]", ""),
        ("()", ""),
        ("{}", ""),  # dict literal is an empty iterator too
        ("frozenset(alpha)", "alpha"),
        ("set(frozenset(alpha))", "alpha"),
        ("list(list(alpha))", "alpha"),
        ("list(sorted(foo))", "foo"),
        ("reversed(reversed(sorted(foo)))", "foo"),
        # TODO: refine `set()` conversions to turn this into
        # a set literal: 'set((x, y))' => '{x, y}'
        ("set(tuple(tuple((x, y))))", "(x, y)"),
    ),
)
def test_iterable_call_under_set_call_is_unwrapped(
    fix_text, set_variant, expression, unwrapped
):
    new_text, _ = fix_text(f"{set_variant}({expression})")
    assert new_text == f"{set_variant}({unwrapped})"


def test_nested_iterable_calls_under_set_work(fix_text):
    new_text, _ = fix_text(
        """\
        set(tuple())
        set(list())
        """
    )
    assert new_text == textwrap.dedent(
        """\
        set()
        set()
        """
    )


@pytest.mark.parametrize(
    "expression, unwrapped",
    (
        ("list(foo)", "foo"),
        ("list(list(foo))", "foo"),
        ("tuple(foo)", "foo"),
        ("tuple(list(tuple(foo)))", "foo"),
    ),
)
def test_order_preserving_iterable_call_under_list_call_is_unwrapped(
    fix_text, expression, unwrapped
):
    new_text, _ = fix_text(f"list({expression})")
    assert new_text == f"list({unwrapped})"


@pytest.mark.parametrize(
    "expression",
    (
        "sorted(foo)",
        "sorted(foo, key=lambda x: x[0])",
    ),
)
def test_reordering_list_call_under_list_call_bubbles_up(fix_text, expression):
    new_text, _ = fix_text(f"list({expression})")
    assert new_text == expression


@pytest.mark.parametrize(
    "expression",
    (
        "list(reversed(foo))",
        "tuple(reversed(foo))",
    ),
)
def test_reversed_ordered_collection_is_preserved(fix_text, expression):
    # confirm that a reverse-order iterator passed into an ordered collection type is
    # preserved
    fix_text(expression, expect_changes=False)



from slyp.codes import ALL_CODES, generate_reference


def test_generated_reference_shows_and_hides_codes_by_hidden():
    doc = "".join(generate_reference())
    for code_def in ALL_CODES:
        if code_def.hidden:
            assert code_def.code not in doc
        else:
            assert code_def.code in doc



import textwrap

import pytest


def test_fixer_inserts_missing_parens_call_arg(fix_text):
    # awkward looking but simple case for paren insertion fixer
    # (black will fix this)
    new_text, _ = fix_text(
        """\
        foo(x="foo"
        "bar")
        """
    )
    assert new_text == textwrap.dedent(
        """\
        foo(x=(
                "foo"
                "bar"
            ))
        """
    )


@pytest.mark.parametrize(
    "disable_comment",
    ("# fmt: off", "#fmt:off ", "#slyp: disable", "# slyp: disable=format "),
)
def test_missing_paren_transform_disabled_with_fmt_off(fix_text, disable_comment):
    fix_text(
        f"""\
        #{disable_comment}
        foo(x="foo"
        "bar")
        """,
        expect_changes=False,
    )


def test_fixer_inserts_missing_parens_call_arg_in_list(fix_text):
    # normal case for paren insertion fixer, in an arg list
    # black fixing will be a no-op because we already handle indents nicely for this
    new_text, _ = fix_text(
        """\
        foo(
            x=1,
            y="foo"
            "bar",
            z=3,
        )
        """
    )
    assert new_text == textwrap.dedent(
        """\
        foo(
            x=1,
            y=(
                "foo"
                "bar"
            ),
            z=3,
        )
        """
    )


@pytest.mark.parametrize("trailer", ("", ",", ",  # comment"))
def test_fixer_inserts_missing_parens_around_dict_element(fix_text, trailer):
    # normal case for paren insertion fixer, in a dict element
    # black fixing will be a no-op because we already handle indents nicely for this
    new_text, _ = fix_text(
        f"""\
        {{
            "foo": "bar "
            "baz"{trailer}
        }}
        """
    )
    assert new_text == textwrap.dedent(
        f"""\
        {{
            "foo": (
                "bar "
                "baz"
            ){trailer}
        }}
        """
    )


@pytest.mark.parametrize(
    "open_delim, close_delim", (("(", ")"), ("[", "]"), ("{", "}"))
)
def test_paren_insertion_in_collection_literal(fix_text, open_delim, close_delim):
    # normal case for paren insertion fixer, in a tuple/list/set
    # black fixing will be a no-op because we already handle indents nicely for this
    new_text, _ = fix_text(
        f"""\
        {open_delim}
            "foo"
            "bar",
            "baz"
        {close_delim}
        """
    )
    assert new_text == textwrap.dedent(
        f"""\
        {open_delim}
            (
                "foo"
                "bar"
            ),
            "baz"
        {close_delim}
        """
    )


@pytest.mark.parametrize(
    "open_delim, close_delim", (("(", ")"), ("[", "]"), ("{", "}"))
)
@pytest.mark.parametrize(
    "outer_open_delim, outer_close_delim", (("(", ")"), ("[", "]"))
)
def test_paren_insertion_in_nested_collection_literal(
    fix_text, open_delim, close_delim, outer_open_delim, outer_close_delim
):
    new_text, _ = fix_text(
        f"""\
        {outer_open_delim}
            {open_delim}
                "foo"
                "bar",
                "baz"
            {close_delim},
            "buzz",
        {outer_close_delim}
        """
    )
    assert new_text == textwrap.dedent(
        f"""\
        {outer_open_delim}
            {open_delim}
                (
                    "foo"
                    "bar"
                ),
                "baz"
            {close_delim},
            "buzz",
        {outer_close_delim}
        """
    )


@pytest.mark.parametrize(
    "open_delim, close_delim", (("(", ")"), ("[", "]"), ("{", "}"))
)
def test_paren_insertion_in_collection_is_noop_on_solo_string(
    fix_text, open_delim, close_delim
):
    fix_text(
        f"""\
        {open_delim}
            "foo"
            "bar"
            "baz"
        {close_delim}
        """,
        expect_changes=False,
    )


@pytest.mark.parametrize(
    "open_delim, close_delim", (("(", ")"), ("[", "]"), ("{", "}"))
)
def test_paren_insertion_in_collection_preserves_and_respaces_comments(
    fix_text, open_delim, close_delim
):
    new_text, _ = fix_text(
        f"""\
        {open_delim}
            "foo"# comment1
            "bar" # comment2
            "baz"  # comment3
            "buzz"   # comment4
            "snork",
            x
        {close_delim}
        """
    )
    assert new_text == textwrap.dedent(
        f"""\
        {open_delim}
            (
                "foo"  # comment1
                "bar"  # comment2
                "baz"  # comment3
                "buzz"  # comment4
                "snork"
            ),
            x
        {close_delim}
        """
    )



from __future__ import annotations

import re

import libcst

from slyp.hashable_file import HashableFile
from slyp.result import Message, Result

from .transformer import SlypTransformer

_DISALBE_RE = re.compile(rb"#\s*((slyp:\s*disable(\=format)?)|(fmt:\s*off))(\s|$)")
_ENABLE_RE = re.compile(rb"#\s*((slyp:\s*enable(\=format)?)|(fmt:\s*on))(\s|$)")


def fix_file(file_obj: HashableFile) -> Result:
    """returns True if no changes were needed"""
    try:
        new_data = _fix_data(file_obj.binary_content)
    # ignore failures to parse and treat these as "unchanged"
    # linting will flag these independently
    except (RecursionError, libcst.ParserSyntaxError, libcst.CSTValidationError):
        return Result(
            success=False,
            messages=[
                Message(f"slyp: failed to parse {file_obj.filename}", verbosity=1)
            ],
        )

    if new_data == file_obj.binary_content:
        if file_obj.is_stdio:
            file_obj.write(file_obj.binary_content)
        return Result(
            messages=[Message(f"slyp: no changes to {file_obj.filename}", verbosity=1)],
            success=True,
        )

    file_obj.write(new_data)
    return Result(messages=[Message(f"slyp: fixed {file_obj.filename}")], success=False)


def _fix_data(content: bytes) -> bytes:
    disabled_line_ranges = _find_disabled_ranges(content)
    raw_tree = libcst.parse_module(content)
    wrapped_tree = libcst.MetadataWrapper(raw_tree)

    tree = wrapped_tree.visit(SlypTransformer(disabled_line_ranges))

    return tree.code.encode(tree.encoding)


def _find_disabled_ranges(content: bytes) -> list[tuple[int, int | float]]:
    start_locations = []
    end_locations = []
    for lineno, line in enumerate(content.split(b"\n")):
        if _DISALBE_RE.search(line):
            start_locations.append(lineno)
        if _ENABLE_RE.search(line):
            end_locations.append(lineno)

    ranges: list[tuple[int, int | float]] = []
    for start in start_locations:
        for end in end_locations:
            if end > start:
                ranges.append((start, end))
                break
        else:
            ranges.append((start, float("inf")))
    return ranges



from __future__ import annotations

import os
import shutil

from slyp.hashable_file import HashableFile

_CACHEDIR = ".slyp_cache"


def _ensure_cachedir(base_cache_dir: str, subdir: str) -> None:
    os.makedirs(base_cache_dir, exist_ok=True)
    os.makedirs(subdir, exist_ok=True)
    gitignore_path = os.path.join(base_cache_dir, ".gitignore")
    if not os.path.exists(gitignore_path):
        with open(gitignore_path, "wb") as fp:
            fp.write(b"*\n")


class PassingFileCache:
    def __init__(
        self,
        *,
        contract_version: str,
        config_id: str,
        base_cache_dir: str = _CACHEDIR,
        cache_dir: str = "passing_files",
    ) -> None:
        self._cache_dir = os.path.join(
            base_cache_dir, f"{cache_dir}_{contract_version}"
        )
        self._config_id = config_id
        _ensure_cachedir(base_cache_dir, self._cache_dir)

    def clear(self) -> None:
        shutil.rmtree(self._cache_dir, ignore_errors=True)

    def _find(self, item: HashableFile) -> str:
        return os.path.join(self._cache_dir, item.sha)

    def __contains__(self, item: HashableFile) -> bool:
        cache_file = self._find(item)
        if not os.path.exists(cache_file):
            return False
        with open(cache_file) as fp:
            data = fp.read()
        return self._config_id in data

    def add(self, item: HashableFile) -> None:
        cache_file = self._find(item)
        if not os.path.exists(cache_file):
            data = self._config_id + "\n"
        else:
            with open(cache_file) as fp:
                data = fp.read().strip()

            if self._config_id in data:  # short-circuit (already recorded)
                return
            else:
                data += f"\n{self._config_id}\n"

        with open(cache_file, "w") as fp:
            fp.write(data)



import textwrap

import pytest


def test_paren_fixer_no_op(fix_text):
    _, res = fix_text(
        "x = 1\n",
        expect_changes=False,
        filename="foo.py",
    )
    assert "slyp: no changes to foo.py" in res.message_strings


def test_paren_fixer_is_no_op_on_needed_parens(fix_text):
    fix_text(
        """\
        a = [(1,), (2, 3)]
        b = foo((x % 2 for x in range(10)), bar)
        c = (1+2)*3
        d = -1 * (-(1+2))
        e = (foo() if x else bar())["baz"]
        f = (await foo())["bar"]
        g = (1).bit_length()
        h = (2.2).is_integer()
        i = (x is y) in truefalsecontainer
        j = lambda: 1, 2
        k = (lambda: 1), 2
        l = (lambda: 1)(2)
        """,
        expect_changes=False,
    )


def test_fix_unnecessary_double_paren_tuple(fix_text):
    new_text, res = fix_text(
        "a = ((1, 2))\n",
        filename="foo.py",
    )

    assert "slyp: fixed foo.py" in res.message_strings
    assert new_text == "a = (1, 2)\n"


def test_fix_unnecessary_many_paren_string(fix_text):
    new_text, _ = fix_text('a = ((((((("foo")))))))\n')
    assert new_text == 'a = "foo"\n'


@pytest.mark.parametrize("layers", (1, 2))
def test_paren_fixer_preserves_innermost_under_splatarg(fix_text, layers):
    new_text, _ = fix_text(
        f"""\
        foo(*{'(' * layers}"a b c".split(){')' * layers})
        """,
        expect_changes=layers > 1,
    )
    assert new_text == 'foo(*("a b c".split()))\n'


def test_multiline_always_allowed(fix_text):
    fix_text(
        """\
        a = (
            (
                (1, 2)
            )
        )
        """,
        expect_changes=False,
    )


def test_names_attributes_and_indexing(fix_text):
    new_text, _ = fix_text(
        """\
        a = (x)
        b = (x.y)
        c = (x).y
        d = x["y"]
        e = (x)["y"]
        f = (x["y"])
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = x
        b = x.y
        c = x.y
        d = x["y"]
        e = x["y"]
        f = x["y"]
        """
    )


def test_collection_types(fix_text):
    new_text, _ = fix_text(
        """\
        a = {1: 2}
        b = ({2: 3})
        c = {x: y for x, y in foo()}
        d = ({x: y for x, y in foo()})
        e = [1, 2]
        f = ([3, 4])
        g = [x for x in foo()]
        h = ([x for x in foo()])
        i = {1}
        j = ({2})
        k = {x for x in foo()}
        l = ({x for x in foo()})
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = {1: 2}
        b = {2: 3}
        c = {x: y for x, y in foo()}
        d = {x: y for x, y in foo()}
        e = [1, 2]
        f = [3, 4]
        g = [x for x in foo()]
        h = [x for x in foo()]
        i = {1}
        j = {2}
        k = {x for x in foo()}
        l = {x for x in foo()}
        """
    )


def test_operators_and_numerics(fix_text):
    new_text, _ = fix_text(
        """\
        a = ((1 + 2)) * (3.2)
        b = ((~ (1 - 2))) * 3 - 2.2
        c = (((a < b) | 1))
        d = a * b
        e = -d
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = (1 + 2) * (3.2)
        b = (~ (1 - 2)) * 3 - 2.2
        c = ((a < b) | 1)
        d = a * b
        e = -d
        """
    )


def test_ellipsis(fix_text):
    new_text, _ = fix_text(
        """\
        a = ((...))
        b = ...
        """
    )
    assert new_text == "a = ...\nb = ...\n"


def test_string_styles(fix_text):
    # note that the ConcatenatedString node will also be fixed to a single string
    new_text, _ = fix_text(
        """\
        x = "foo"
        a = ("foo")
        b = (("foo " "bar"))
        c = (((f"baz {a}")))
        """
    )
    assert new_text == textwrap.dedent(
        """\
        x = "foo"
        a = "foo"
        b = "foo bar"
        c = f"baz {a}"
        """
    )


def test_lambdas_and_yields_with_many_parens(fix_text):
    new_text, _ = fix_text(
        """\
        a = ((lambda: 1))
        b = (lambda: ((1)))
        c = lambda: (yield)
        d = lambda: ((yield))
        e = lambda: (yield from foo())
        f = lambda: ((yield from foo()))
        g = (lambda: ((yield from foo())))
        h = ((lambda: ((yield from foo()))))
        i = ((((lambda: ((yield from foo()))))))
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = (lambda: 1)
        b = (lambda: (1))
        c = lambda: (yield)
        d = lambda: (yield)
        e = lambda: (yield from foo())
        f = lambda: (yield from foo())
        g = (lambda: (yield from foo()))
        h = (lambda: (yield from foo()))
        i = (lambda: (yield from foo()))
        """
    )


def test_generator_expressions(fix_text):
    new_text, _ = fix_text(
        """\
        a = foo(x for x in bar())
        b = foo((x for x in bar()), baz())
        c = foo(((x for x in bar())), baz())
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = foo(x for x in bar())
        b = foo((x for x in bar()), baz())
        c = foo((x for x in bar()), baz())
        """
    )


def test_if_expressions(fix_text):
    new_text, _ = fix_text(
        """\
        a = (1 if x else 2) | 3
        b = 1 if x else 2 & a
        c = ((1 if a else 2)) or b
        """,
    )
    assert new_text == textwrap.dedent(
        """\
        a = (1 if x else 2) | 3
        b = 1 if x else 2 & a
        c = (1 if a else 2) or b
        """,
    )


def test_await_safe_use_for_precedence(fix_text):
    new_text, _ = fix_text(
        """\
        a = (await foo() if x else bar())["baz"]
        b = ((await foo() if x else bar()))
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = (await foo() if x else bar())["baz"]
        b = (await foo() if x else bar())
        """
    )


def test_match_with_paren_gets_space_inserted(fix_text):
    new_text, _ = fix_text(
        """\
        match(x):
            case (1, 2):
                pass
            case _:
                pass
        """
    )
    assert new_text == textwrap.dedent(
        """\
        match x:
            case (1, 2):
                pass
            case _:
                pass
        """
    )


def test_with_with_paren_gets_space_inserted(fix_text):
    new_text, _ = fix_text(
        """\
        with(x()):
            pass
        """
    )
    assert new_text == textwrap.dedent(
        """\
        with x():
            pass
        """
    )


def test_if_with_paren_gets_space_inserted(fix_text):
    new_text, _ = fix_text(
        """\
        if(x()):
            pass
        """
    )

    # after the first round of fixing, space is inserted but the parens are preserved
    # this ensures that we never generate an invalid node which libcst 1.5.0+ will
    # reject
    assert new_text == textwrap.dedent(
        """\
        if (x()):
            pass
        """
    )

    # confirm that a second round of fixing is sufficient and gets to the desired
    # end-result
    new_text, _ = fix_text(new_text)

    assert new_text == textwrap.dedent(
        """\
        if x():
            pass
        """
    )


def test_import_from_gets_space_inserted(fix_text):
    new_text, _ = fix_text(
        """\
        from foo import(bar, baz)
        """
    )
    assert new_text == "from foo import bar, baz\n"


def test_unary_op_preserves_parens(fix_text):
    fix_text(
        """\
        foo(~(bar.baz()))
        """,
        expect_changes=False,
    )



import textwrap

import pytest


def test_simple_string_concat_with_matching_characteristics(fix_text):
    new_text, _ = fix_text(
        """\
        a = "foo " "bar"
        b = 'foo ' 'bar'
        c = r"foo " r"bar"
        d = u"foo " u"bar"
        e = b"foo " b"bar"
        f = br"foo " br"bar"
        g = rb"foo " rb"bar"
        """
    )
    assert new_text == textwrap.dedent(
        """\
        a = "foo bar"
        b = 'foo bar'
        c = r"foo bar"
        d = u"foo bar"
        e = b"foo bar"
        f = br"foo bar"
        g = rb"foo bar"
        """
    )


def test_fmt_toggle_can_reenable_fixing(fix_text):
    new_text, _ = fix_text(
        """\
        # slyp: disable
        # fmt: on
        a = "foo " "bar"
        """,
    )
    assert new_text == textwrap.dedent(
        """\
        # slyp: disable
        # fmt: on
        a = "foo bar"
        """
    )


def test_quote_style_mismatch_suppresses_fixing(fix_text):
    fix_text(
        """\
        a = "foo " 'bar'
        """,
        expect_changes=False,
    )


@pytest.mark.parametrize(
    "prefix_a, prefix_b",
    [
        (prefix_a, prefix_b)
        for prefix_a in ("", "r", "u", "b", "br", "rb")
        for prefix_b in ("", "r", "u", "b", "br", "rb")
        # get the cases where they differ
        if prefix_a != prefix_b and
        # but not the ones where a str and bytes are concatenated
        ("b" in prefix_a) == ("b" in prefix_b)
        # and not the special case of "br" and "rb", which are really the same
        and {prefix_a, prefix_b} != {"br", "rb"}
    ],
)
def test_prefix_mismatch_suppresses_fixing(fix_text, prefix_a, prefix_b):
    fix_text(
        f"""\
        a = {prefix_a}"foo " {prefix_b}"bar"
        """,
        expect_changes=False,
    )


def test_prefix_mismatch_is_ignored_in_special_cases(fix_text):
    new_text, _ = fix_text(
        """\
        a = br"foo " rb"bar"
        b = rb"foo " br"bar"
        """,
    )
    assert new_text == textwrap.dedent(
        """\
        a = br"foo bar"
        b = rb"foo bar"
        """
    )


def test_double_concat(fix_text):
    new_text, _ = fix_text(
        """\
        x = "foo " "bar " "baz"
        """,
    )
    assert new_text == textwrap.dedent(
        """\
        x = "foo bar baz"
        """
    )


def test_concat_fstrings(fix_text):
    new_text, _ = fix_text(
        """\
        x = f"{foo}" f"{bar}"
        y = fr"{foo}" Fr"{bar}"
        """,
    )
    assert new_text == textwrap.dedent(
        """\
        x = f"{foo}{bar}"
        y = fr"{foo}{bar}"
        """
    )


def test_concat_fstring_with_simple(fix_text):
    new_text, _ = fix_text(
        """\
        x = f"{foo}" "bar"
        y = "foo" f"{bar}"
        p = fr"{foo}" r"bar"
        q = r"foo" rf"{bar}"
        """,
    )
    assert new_text == textwrap.dedent(
        """\
        x = f"{foo}bar"
        y = f"foo{bar}"
        p = fr"{foo}bar"
        q = rf"foo{bar}"
        """
    )


def test_concat_fstring_skips_mismatched_quotes(fix_text):
    fix_text(
        """\
        x = f'{foo}' "bar"
        y = 'foo' f"{bar}"
        """,
        expect_changes=False,
    )


def test_concat_fstring_skips_if_curly_braces_present(fix_text):
    fix_text(
        """\
        x = f'{foo}' "bar{}"
        x = 'foo{}' f"{bar}"
        """,
        expect_changes=False,
    )



import pytest


@pytest.mark.parametrize("expr", ("if", "try", "ifexpr", "ifexpr-in-if-test"))
def test_check_captures_w200(check_text, expr):
    if expr == "if":
        content = """\
            def foo():
                if bar():
                    return baz(quux("snork"))
                else:
                    return baz(quux("snork"))
        """
    elif expr == "try":
        content = """\
            def foo():
                try:
                    return baz(quux("snork"))
                except:
                    return baz(quux("snork"))
        """
    elif expr == "ifexpr":
        content = """\
            def foo():
                return baz(quux("snork")) if buzz() else baz(quux("snork"))
        """
    elif expr == "ifexpr-in-if-test":
        content = """\
            def foo():
                if baz(quux("snork")) if buzz() else baz(quux("snork")):
                    return 1
                return 0
        """
    else:
        raise NotImplementedError

    res = check_text(content, filename="foo.py")

    assert not res.success
    assert (
        "foo.py:2: two AST branches have identical contents (W200)"
        in res.message_strings
    )


def test_check_captures_w201(check_text):
    res = check_text(
        """\
        def foo():
            if bar():
                return baz()
            else:
                return baz()
        """,
        filename="foo.py",
    )

    assert not res.success
    assert (
        "foo.py:2: two AST branches have identical trivial contents (W201)"
        in res.message_strings
    )


def test_check_captures_w202(check_text):
    res = check_text(
        """\
        def foo():
            if bar():
                return baz("snork", flip="flop")
            elif qux():
                return quux()
            else:
                return baz("snork", flip="flop")
        """,
        filename="foo.py",
    )

    assert not res.success
    assert (
        "foo.py:2: two non-adjacent AST branches have identical contents (W202)"
        in res.message_strings
    )


def test_check_captures_w203(check_text):
    res = check_text(
        """\
        def foo():
            if bar():
                return baz("snork")
            elif qux():
                return quux()
            else:
                return baz("snork")
        """,
        filename="foo.py",
    )

    assert not res.success
    assert (
        "foo.py:2: two non-adjacent AST branches have identical trivial contents (W203)"
        in res.message_strings
    )


def test_captures_nested_w200s(check_text):
    res = check_text(
        """\
        def foo():
            if bar():
                return baz(quux("snork"))
            elif bar2():
                return baz(quux("snork"))
            else:
                try:
                    return baz(snork("quux"))
                except:
                    return baz(snork("quux"))
        """,
        filename="foo.py",
    )

    assert not res.success
    assert (
        "foo.py:2: two AST branches have identical contents (W200)"
        in res.message_strings
    )
    assert (
        "foo.py:7: two AST branches have identical contents (W200)"
        in res.message_strings
    )



from __future__ import annotations

import typing as t

import libcst
import libcst.matchers

# an __init__ definition missing the return type annotation
MISSING_RETURN_ANNOTATION_INIT_MATCHER = libcst.matchers.FunctionDef(
    name=libcst.matchers.Name(value="__init__"), returns=None
)

# SimpleWhitespace defines whitespace as seen between tokens in most contexts
# it can contain newlines *if* there is a preceding backslash escape
SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER = libcst.matchers.SimpleWhitespace(
    value=libcst.matchers.MatchIfTrue(lambda x: "\n" not in x)
)

# define a concatenated string over multiple lines, with no parens wrapping it
# restrict this to only match when there is no comment between the lines
UNPARENTHESIZED_MULTILINE_CONCATENATED_STRING_MATCHER = libcst.matchers.ConcatenatedString(  # noqa: E501
    whitespace_between=(
        libcst.matchers.SimpleWhitespace(
            value=libcst.matchers.MatchIfTrue(lambda x: "\n" in x)
        )
        | libcst.matchers.ParenthesizedWhitespace()
    ),
    # empty list means that there are no parens
    # because `lpar` and `rpar` are `Sequence[Paren]` types
    lpar=[],
    rpar=[],
)

# check for 'unparenthesized multiline string concat in container'
#
# this case is multiline str concat inside of a container type, containing other
# elements
#
# the main scenario is "accidental multiline string", where a collection contains
# string literals, but is missing one comma
#
#   x = [
#       "foo "
#       "bar",
#       "baz",
#   ]
#
# however, a similar argument applies to any mixed iterable, e.g.
#
#   x = (
#       "foo "
#       "bar",
#       baz(),
#   ]
UNPARENTHESIZED_CONCAT_ELEMENT_MATCHER = libcst.matchers.Element(
    value=UNPARENTHESIZED_MULTILINE_CONCATENATED_STRING_MATCHER
)

# this allows us to do faster matching for W103 by checking the whole list with
# a single matcher conditional
UNPARENTHESIZED_CONCAT_IN_ELEMENT_LIST_MATCHER = libcst.matchers.OneOf(
    [
        libcst.matchers.AtLeastN(n=1),
        UNPARENTHESIZED_CONCAT_ELEMENT_MATCHER,
        libcst.matchers.ZeroOrMore(),
    ],
    [
        libcst.matchers.ZeroOrMore(),
        UNPARENTHESIZED_CONCAT_ELEMENT_MATCHER,
        libcst.matchers.AtLeastN(n=1),
    ],
)

PARENT_NODE_REQUIRES_INNERMOST_PARENS_MATCHER = libcst.matchers.OneOf(
    libcst.matchers.Arg(star=libcst.matchers.MatchIfTrue(lambda x: "*" in x)),
    libcst.matchers.UnaryOperation(),
    libcst.matchers.If(whitespace_before_test=libcst.matchers.SimpleWhitespace("")),
    libcst.matchers.While(whitespace_after_while=libcst.matchers.SimpleWhitespace("")),
)

ParenFixNodeTypes = t.Union[
    libcst.Name,
    libcst.Attribute,
    libcst.Subscript,
    libcst.Call,
    libcst.Dict,
    libcst.DictComp,
    libcst.List,
    libcst.ListComp,
    libcst.Set,
    libcst.SetComp,
    libcst.Tuple,
    libcst.UnaryOperation,
    libcst.BinaryOperation,
    libcst.Comparison,
    libcst.Ellipsis,
    libcst.Integer,
    libcst.Float,
    libcst.SimpleString,
    libcst.ConcatenatedString,
    libcst.FormattedString,
    libcst.Lambda,
    libcst.Yield,
    libcst.GeneratorExp,
    libcst.IfExp,
    libcst.Await,
]
PFN = t.TypeVar("PFN", bound=ParenFixNodeTypes)


CollectionLiteralNode = t.TypeVar(
    "CollectionLiteralNode", bound=t.Union[libcst.Tuple, libcst.List, libcst.Set]
)


class SlypTransformer(libcst.CSTTransformer):
    METADATA_DEPENDENCIES = (
        libcst.metadata.PositionProvider,
        libcst.metadata.ParentNodeProvider,
    )

    def __init__(self, disabled_line_ranges: list[tuple[int, int | float]]) -> None:
        self.disabled_line_ranges = disabled_line_ranges

    def on_leave(
        self, original_node: libcst.CSTNodeT, updated_node: libcst.CSTNodeT
    ) -> (
        libcst.CSTNodeT
        | libcst.RemovalSentinel
        | libcst.FlattenSentinel[libcst.CSTNodeT]
    ):
        new_updated_node = super().on_leave(original_node, updated_node)
        if new_updated_node != updated_node:
            # if the node has been updated, check if disabled
            if not self._node_is_disabled(original_node):
                return new_updated_node
        return updated_node

    def _node_is_disabled(self, node: libcst.CSTNode) -> bool:
        if not self.disabled_line_ranges:
            return False
        start_line = self.get_metadata(
            libcst.metadata.PositionProvider, node
        ).start.line
        for start, end in self.disabled_line_ranges:
            if start <= start_line < end:
                return True
        return False

    def _singular_parens_are_same_line(
        self, node: libcst.With | libcst.ImportFrom
    ) -> bool:
        if node.lpar is not None and libcst.matchers.matches(
            node.lpar, libcst.matchers.LeftParen()
        ):
            lpar_line = self.get_metadata(
                libcst.metadata.PositionProvider, node.lpar  # type: ignore[arg-type]
            ).start.line
            rpar_line = self.get_metadata(
                libcst.metadata.PositionProvider, node.rpar  # type: ignore[arg-type]
            ).start.line
            return lpar_line == rpar_line

        return False

    def _how_many_parens_same_line(self, node: PFN) -> int:
        """
        Determine how many parens are on the same line for a node.

        Start by finding the "max offset" into the paren lists where the lpar and rpar
        are on the same line. The range is [-1 , len(paren_list)] ; then add one.
        """
        max_offset = -1
        for offset in range(len(node.lpar)):
            lpar_line = self.get_metadata(
                libcst.metadata.PositionProvider, node.lpar[-(offset + 1)]
            ).start.line
            rpar_line = self.get_metadata(
                libcst.metadata.PositionProvider, node.rpar[offset]
            ).start.line

            if lpar_line != rpar_line:
                break
            max_offset = offset
        return max_offset + 1

    def modify_parenthesized_node(
        self,
        original_node: ParenFixNodeTypes,
        updated_node: PFN,
        *,
        preserve_innermost: bool = False,
    ) -> PFN:
        num_parens_to_unwrap: int | None = None
        if not preserve_innermost:
            # **optimization**
            #
            # if there is a parent node to check, pick up on the number of parens to
            # unwrap first -- this saves the work of accessing the parent node when it's
            # avoidable and we can short-circuit
            #
            # based on timings, accessing parent nodes seems slower than finding paren
            # linenos
            num_parens_to_unwrap = self._how_many_parens_same_line(original_node)
            if num_parens_to_unwrap <= 0:
                return updated_node

            parent = self.get_metadata(
                libcst.metadata.ParentNodeProvider, original_node
            )
            # check if the parent of the node is *-expansion of an arg
            # or an `if` or `while` without whitespace before the condition
            if libcst.matchers.matches(
                parent, PARENT_NODE_REQUIRES_INNERMOST_PARENS_MATCHER
            ):
                preserve_innermost = True

        if preserve_innermost and len(original_node.lpar) == 1:
            return updated_node

        if num_parens_to_unwrap is None:
            num_parens_to_unwrap = self._how_many_parens_same_line(original_node)
        if preserve_innermost:
            num_parens_to_unwrap -= 1
        if num_parens_to_unwrap <= 0:
            return updated_node
        return updated_node.with_changes(  # type: ignore[return-value]
            lpar=updated_node.lpar[:-num_parens_to_unwrap],
            rpar=updated_node.rpar[num_parens_to_unwrap:],
        )

    def refold_and_parenthesize_str_concat_node(
        self,
        node: libcst.ConcatenatedString,
    ) -> libcst.ConcatenatedString:
        """given a concatenated string node, add parens and "refold" the whitespace"""
        # build a list of concatenated string nodes (unroll the recursive structure)
        # including the final non-ConcatenatedString node
        concat_nodes: list[libcst.BaseString] = []
        cur: libcst.BaseString = node
        while isinstance(cur, libcst.ConcatenatedString):
            concat_nodes.append(cur)
            cur = cur.right
        concat_nodes.append(cur)

        base_indent = max(_find_max_indent_of_concat_node(concat_nodes), 4) + 4

        # walk the list in reverse, updating the whitespace between the nodes and
        # "clipping them back together" (hence the reversal here, maintaining the
        # right-heavy structure of the original nodes)
        for idx in range(len(concat_nodes) - 2, -1, -1):
            cur = concat_nodes[idx]
            comment = None
            if isinstance(
                cur.whitespace_between,  # type: ignore[attr-defined]
                libcst.ParenthesizedWhitespace,
            ):
                comment = cur.whitespace_between.first_line.comment  # type: ignore[attr-defined]  # noqa: E501
            concat_nodes[idx] = cur.with_changes(
                whitespace_between=_make_paren_whitespace(
                    " " * base_indent, comment=comment
                ),
                right=concat_nodes[idx + 1],
            )

        return concat_nodes[0].with_changes(  # type: ignore[return-value]
            lpar=[
                libcst.LeftParen(
                    whitespace_after=_make_paren_whitespace(" " * base_indent)
                )
            ],
            rpar=[
                libcst.RightParen(
                    whitespace_before=_make_paren_whitespace(" " * (base_indent - 4))
                )
            ],
        )

    def refold_element_list(self, node: CollectionLiteralNode) -> CollectionLiteralNode:
        return node.with_changes(  # type: ignore[return-value]
            elements=[
                (
                    e.with_changes(
                        value=self.refold_and_parenthesize_str_concat_node(
                            e.value  # type: ignore[arg-type]
                        )
                    )
                    if libcst.matchers.matches(
                        e, UNPARENTHESIZED_CONCAT_ELEMENT_MATCHER
                    )
                    else e
                )
                for e in node.elements
            ]
        )

    # identifiers, attributes, lookups, and calls

    def leave_Name(
        self, original_node: libcst.Name, updated_node: libcst.Name
    ) -> libcst.Name:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Attribute(
        self, original_node: libcst.Attribute, updated_node: libcst.Attribute
    ) -> libcst.Attribute:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Subscript(
        self, original_node: libcst.Subscript, updated_node: libcst.Subscript
    ) -> libcst.Subscript:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Call(
        self, original_node: libcst.Call, updated_node: libcst.Call
    ) -> libcst.BaseExpression:
        if original_node.lpar:
            updated_node = self.modify_parenthesized_node(original_node, updated_node)

        if not libcst.matchers.matches(
            updated_node,
            libcst.matchers.Call(
                func=libcst.matchers.Name("dict")
                | libcst.matchers.Name("list")
                | libcst.matchers.Name("tuple")
                | libcst.matchers.Name("set")
                | libcst.matchers.Name("frozenset")
            ),
        ):
            return updated_node

        func_name: str = original_node.func.value  # type: ignore[attr-defined]
        if func_name == "dict":
            return self._fix_dict_call(original_node, updated_node)
        elif func_name == "list":
            return self._fix_list_call(original_node, updated_node)
        elif func_name == "tuple":
            return self._fix_tuple_call(original_node, updated_node)
        elif func_name in ("set", "frozenset"):
            return self._fix_set_call(original_node, updated_node)
        else:
            return updated_node

    def _fix_dict_call(
        self, original_node: libcst.Call, updated_node: libcst.Call
    ) -> libcst.Call | libcst.Dict:
        # match a 'dict()' call whose args can unpack to a dict literal
        if not libcst.matchers.matches(
            updated_node,
            libcst.matchers.Call(
                func=libcst.matchers.Name("dict"),
                args=[
                    libcst.matchers.ZeroOrMore(
                        libcst.matchers.Arg(star="**")
                        | libcst.matchers.Arg(keyword=libcst.matchers.Name())
                    )
                ],
            ),
        ):
            return updated_node

        return libcst.Dict(
            elements=[_convert_dict_element(arg) for arg in updated_node.args],
            lpar=updated_node.lpar,
            rpar=updated_node.rpar,
            lbrace=libcst.LeftCurlyBrace(
                whitespace_after=updated_node.whitespace_before_args
            ),
            rbrace=libcst.RightCurlyBrace(
                whitespace_before=_last_arg_whitespace(updated_node)
            ),
        )

    def _fix_list_call(
        self, original_node: libcst.Call, updated_node: libcst.Call
    ) -> libcst.Call | libcst.List | libcst.ListComp:
        # 'list(X(...))' where `X` is a known function call which produces a list
        # already
        # OR where `X` is `tuple`
        # (which is interesting because it does not transform order)
        #
        # this fix is run in a loop to repeatedly unnest expressions
        # thus fixing cases like `list(list(sorted(foo))`
        while libcst.matchers.matches(
            updated_node,
            libcst.matchers.Call(
                func=libcst.matchers.Name("list"),
                args=[
                    libcst.matchers.Arg(
                        star="",
                        keyword=None,
                        value=libcst.matchers.Call(
                            func=libcst.matchers.Name("list")
                            | libcst.matchers.Name("sorted")
                            | libcst.matchers.Name("tuple")
                        ),
                    )
                ],
            ),
        ):
            arg0: libcst.Call = updated_node.args[0].value  # type: ignore[assignment]
            if arg0.func.value == "tuple":  # type: ignore[attr-defined]
                updated_node = updated_node.with_changes(args=arg0.args)
            else:
                updated_node = updated_node.with_changes(func=arg0.func, args=arg0.args)

        # a 'list()' call with no arguments
        if libcst.matchers.matches(
            updated_node,
            libcst.matchers.Call(
                func=libcst.matchers.Name("list"),
                args=[],
            ),
        ):
            return libcst.List(
                elements=[], lpar=updated_node.lpar, rpar=updated_node.rpar
            )

        # a 'list()' call whose only argument is a generator expression
        if libcst.matchers.matches(
            updated_node,
            libcst.matchers.Call(
                func=libcst.matchers.Name("list"),
                args=[
                    libcst.matchers.Arg(
                        star="", keyword=None, value=libcst.matchers.GeneratorExp()
                    )
                ],
            ),
        ):
            genexp: libcst.GeneratorExp = updated_node.args[
                0
            ].value  # type: ignore[assignment]
            return libcst.ListComp(
                elt=genexp.elt,
                for_in=genexp.for_in,
                lpar=updated_node.lpar,
                rpar=updated_node.rpar,
            )

        return updated_node

    def _fix_tuple_call(
        self, original_node: libcst.Call, updated_node: libcst.Call
    ) -> libcst.Call | libcst.Tuple:
        # a 'list()' call with no arguments
        if libcst.matchers.matches(
            updated_node,
            libcst.matchers.Call(
                func=libcst.matchers.Name("tuple"),
                args=[],
            ),
        ):
            lpar = updated_node.lpar if updated_node.lpar else [libcst.LeftParen()]
            rpar = updated_node.rpar if updated_node.rpar else [libcst.RightParen()]
            return libcst.Tuple(elements=[], lpar=lpar, rpar=rpar)

        return updated_node

    def _fix_set_call(
        self, original_node: libcst.Call, updated_node: libcst.Call
    ) -> libcst.Call | libcst.SetComp:
        """accepts `set()` and `frozenset()` calls"""
        # 'set(X(...))' where `X` is a known function call which produces an iterator
        # or 'frozenset(X(...))'
        #
        # this fix is run in a loop to repeatedly unnest expressions
        # thus fixing cases like `set(reversed(sorted(foo)))`
        while libcst.matchers.matches(
            updated_node,
            libcst.matchers.Call(
                func=libcst.matchers.Name("set") | libcst.matchers.Name("frozenset"),
                args=[
                    libcst.matchers.Arg(
                        star="",
                        keyword=None,
                        value=libcst.matchers.Call(
                            func=libcst.matchers.Name("sorted")
                            | libcst.matchers.Name("reversed")
                            | libcst.matchers.Name("list")
                            | libcst.matchers.Name("set")
                            | libcst.matchers.Name("frozenset")
                            | libcst.matchers.Name("tuple")
                        ),
                    )
                ],
            ),
        ):
            arg0: libcst.Call = updated_node.args[0].value  # type: ignore[assignment]
            if not arg0.args:
                # if we are seeing no args, like `set(set())`, that's just `set()`
                updated_node = updated_node.with_changes(args=[])
            else:
                # otherwise, we only need to preserve the first arg
                # `set(sorted(X, reversed=True))` => `set(X)`
                arg0_arg0: libcst.Arg = arg0.args[0]
                updated_node = updated_node.with_changes(
                    # use a new Arg to discard any comma and whitespace
                    args=[libcst.Arg(value=arg0_arg0.value)],
                )

        # a 'set()' call whose only argument is a generator expression
        if libcst.matchers.matches(
            updated_node,
            libcst.matchers.Call(
                func=libcst.matchers.Name("set"),
                args=[
                    libcst.matchers.Arg(
                        star="", keyword=None, value=libcst.matchers.GeneratorExp()
                    )
                ],
            ),
        ):
            genexp: libcst.GeneratorExp = updated_node.args[
                0
            ].value  # type: ignore[assignment]
            return libcst.SetComp(
                elt=genexp.elt,
                for_in=genexp.for_in,
                lpar=updated_node.lpar,
                rpar=updated_node.rpar,
            )

        # a 'set()' call whose only argument is an empty collection literal should
        # be an empty call
        # (or a 'frozenset()' call)
        #
        # e.g.  'set([])' => 'set()'
        if libcst.matchers.matches(
            updated_node,
            libcst.matchers.Call(
                func=libcst.matchers.Name("set") | libcst.matchers.Name("frozenset"),
                args=[
                    libcst.matchers.Arg(
                        star="",
                        keyword=None,
                        value=libcst.matchers.Tuple(elements=[])
                        | libcst.matchers.List(elements=[])
                        | libcst.matchers.Dict(elements=[]),
                    )
                ],
            ),
        ):
            return updated_node.with_changes(args=[])

        return updated_node

    def leave_Arg(
        self, original_node: libcst.Arg, updated_node: libcst.Arg
    ) -> libcst.Arg:
        # the scenario here is that the string concat happens across multiple
        # lines, but without parenthesization
        # with a keyword argument getting the value, this is valid but results
        # in a string with different physical indentation levels in the file
        # e.g.
        #   foo(
        #       bar="alpha "
        #       "beta"
        #   )
        #
        # a classic case where this arises naturally is `help=...` for
        # argparse and click, where the help string is often slightly too long
        # for a single line, and easy to "break in two" without adding parens to
        # force the whole block to indent
        if libcst.matchers.matches(
            original_node,
            libcst.matchers.Arg(
                keyword=libcst.matchers.Name(),
                value=UNPARENTHESIZED_MULTILINE_CONCATENATED_STRING_MATCHER,
            ),
        ):
            return updated_node.with_changes(
                value=self.refold_and_parenthesize_str_concat_node(
                    updated_node.value  # type: ignore[arg-type]
                )
            )
        return updated_node

    def leave_DictElement(
        self, original_node: libcst.DictElement, updated_node: libcst.DictElement
    ) -> libcst.DictElement:
        if original_node.whitespace_after_colon.empty:
            updated_node = updated_node.with_changes(
                whitespace_after_colon=libcst.SimpleWhitespace(" ")
            )
        if libcst.matchers.matches(
            original_node.value,
            UNPARENTHESIZED_MULTILINE_CONCATENATED_STRING_MATCHER,
        ):
            updated_node = updated_node.with_changes(
                value=self.refold_and_parenthesize_str_concat_node(
                    updated_node.value  # type: ignore[arg-type]
                )
            )
        return updated_node

    # collections

    def leave_Dict(
        self, original_node: libcst.Dict, updated_node: libcst.Dict
    ) -> libcst.Dict:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_DictComp(
        self, original_node: libcst.DictComp, updated_node: libcst.DictComp
    ) -> libcst.DictComp:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_List(
        self, original_node: libcst.List, updated_node: libcst.List
    ) -> libcst.List:
        if original_node.lpar:
            updated_node = self.modify_parenthesized_node(original_node, updated_node)
        if libcst.matchers.matches(
            original_node,
            libcst.matchers.List(
                elements=UNPARENTHESIZED_CONCAT_IN_ELEMENT_LIST_MATCHER  # type: ignore[arg-type]  # noqa: E501
            ),
        ):
            updated_node = self.refold_element_list(updated_node)
        return updated_node

    def leave_ListComp(
        self, original_node: libcst.ListComp, updated_node: libcst.ListComp
    ) -> libcst.ListComp:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Set(
        self, original_node: libcst.Set, updated_node: libcst.Set
    ) -> libcst.Set:
        if original_node.lpar:
            updated_node = self.modify_parenthesized_node(original_node, updated_node)
        if libcst.matchers.matches(
            original_node,
            libcst.matchers.Set(
                elements=UNPARENTHESIZED_CONCAT_IN_ELEMENT_LIST_MATCHER  # type: ignore[arg-type]  # noqa: E501
            ),
        ):
            updated_node = self.refold_element_list(updated_node)
        return updated_node

    def leave_SetComp(
        self, original_node: libcst.SetComp, updated_node: libcst.SetComp
    ) -> libcst.SetComp:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Tuple(
        self, original_node: libcst.Tuple, updated_node: libcst.Tuple
    ) -> libcst.Tuple:
        # list of tuples:
        #   [(1, 2), (3, 4)] != [1, 2, 3, 4]
        if len(original_node.lpar) >= 2:
            updated_node = self.modify_parenthesized_node(
                original_node,
                updated_node,
                preserve_innermost=True,
            )
        if libcst.matchers.matches(
            original_node,
            libcst.matchers.Tuple(
                elements=UNPARENTHESIZED_CONCAT_IN_ELEMENT_LIST_MATCHER  # type: ignore[arg-type]  # noqa: E501
            ),
        ):
            updated_node = self.refold_element_list(updated_node)
        return updated_node

    # operators

    def leave_UnaryOperation(
        self, original_node: libcst.UnaryOperation, updated_node: libcst.UnaryOperation
    ) -> libcst.UnaryOperation:
        # parens define precedence
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_BinaryOperation(
        self,
        original_node: libcst.BinaryOperation,
        updated_node: libcst.BinaryOperation,
    ) -> libcst.BinaryOperation:
        # parens define precedence
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_Comparison(
        self, original_node: libcst.Comparison, updated_node: libcst.Comparison
    ) -> libcst.Comparison:
        # parens define precedence
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    # literals, constants, and strings

    def leave_Ellipsis(
        self, original_node: libcst.Ellipsis, updated_node: libcst.Ellipsis
    ) -> libcst.Ellipsis:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Integer(
        self,
        original_node: libcst.Integer,
        updated_node: libcst.Integer,
    ) -> libcst.Integer:
        # direct access to attributes of integers and floats requires parens:
        #   (1).bit_length()
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_Float(
        self,
        original_node: libcst.Float,
        updated_node: libcst.Float,
    ) -> libcst.Float:
        # direct access to attributes of integers and floats requires parens:
        #   (1).bit_length()
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_SimpleString(
        self, original_node: libcst.SimpleString, updated_node: libcst.SimpleString
    ) -> libcst.SimpleString:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_ConcatenatedString(
        self,
        original_node: libcst.ConcatenatedString,
        updated_node: libcst.ConcatenatedString,
    ) -> libcst.ConcatenatedString | libcst.SimpleString | libcst.FormattedString:
        new_node = updated_node
        # if the node is parenthesized, potentially strip parens
        if original_node.lpar:
            new_node = self.modify_parenthesized_node(original_node, updated_node)

        left: libcst.SimpleString
        right: libcst.SimpleString
        left_f: libcst.FormattedString
        right_f: libcst.FormattedString

        # if the node is a pair of simple strings with no newline between them,
        # this may be a chance to join them into a single string node
        if libcst.matchers.matches(
            new_node,
            libcst.matchers.ConcatenatedString(
                whitespace_between=SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER,
                left=libcst.matchers.SimpleString(),
                right=libcst.matchers.SimpleString(),
            ),
        ):
            # check that the left and right match in their prefix and quote characters,
            # and forbid this change (for now) when one of the two is a multiline string
            # (TODO: consider when it might be safe to join multiline strings)
            #
            # the restriction against differing quote characters allows us to avoid any
            # attempt to manipulate unescaped quotes in the string values
            # having done this verification, join the strings into a single node,
            # preserving the prefix and quote style
            left = new_node.left  # type: ignore[assignment]
            right = new_node.right  # type: ignore[assignment]
            if (
                (
                    left.prefix == right.prefix
                    or {left.prefix, right.prefix}.issubset({"br", "rb"})
                )
                and left.quote == right.quote
                and left.quote in {"'", '"'}
            ):
                return libcst.SimpleString(
                    lpar=new_node.lpar,
                    rpar=new_node.rpar,
                    value=left.prefix
                    + left.quote
                    + left.raw_value
                    + right.raw_value
                    + left.quote,
                )
        # two f-strings on a line, like
        #   f"{foo} " f"{bar}"
        # which can merge to
        #   f"{foo} {bar}"
        elif libcst.matchers.matches(
            new_node,
            libcst.matchers.ConcatenatedString(
                whitespace_between=SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER,
                left=libcst.matchers.FormattedString(),
                right=libcst.matchers.FormattedString(),
            ),
        ):
            left_f = new_node.left  # type: ignore[assignment]
            right_f = new_node.right  # type: ignore[assignment]

            if (
                (
                    left_f.prefix == right_f.prefix
                    or {left_f.prefix, right_f.prefix}.issubset({"fr", "rf"})
                )
                and left_f.quote == right_f.quote
                and left_f.quote in {"'", '"'}
            ):
                return libcst.FormattedString(
                    lpar=new_node.lpar,
                    rpar=new_node.rpar,
                    parts=(left_f.parts + right_f.parts),  # type: ignore[operator]
                    start=left_f.start,
                    end=right_f.end,
                )
        # if it's a format string with a string that has no curly-braces (left)
        # then we can join it, e.g.
        #   "foo " f"{bar}"  ->  f"foo {bar}"
        elif libcst.matchers.matches(
            new_node,
            libcst.matchers.ConcatenatedString(
                whitespace_between=SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER,
                left=libcst.matchers.SimpleString(),
                right=libcst.matchers.FormattedString(),
            ),
        ):
            left = new_node.left  # type: ignore[assignment]
            right_f = new_node.right  # type: ignore[assignment]
            if (
                (
                    (left.prefix, right_f.prefix) == ("", "f")
                    or {left.prefix, right_f.prefix}.issubset({"r", "rf", "fr"})
                )
                and left.quote == right_f.quote
                and left.quote in {"'", '"'}
                and "{" not in left.raw_value
                and "}" not in left.raw_value
            ):
                return libcst.FormattedString(
                    lpar=new_node.lpar,
                    rpar=new_node.rpar,
                    parts=(
                        [libcst.FormattedStringText(value=left.raw_value)]
                        + list(right_f.parts)
                    ),
                    start=right_f.start,
                    end=right_f.end,
                )
        # same as the above, right-hand side; e.g.
        #   f"{foo} " "bar"  ->  f"{foo} bar"
        elif libcst.matchers.matches(
            new_node,
            libcst.matchers.ConcatenatedString(
                whitespace_between=SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER,
                left=libcst.matchers.FormattedString(),
                right=libcst.matchers.SimpleString(),
            ),
        ):
            left_f = new_node.left  # type: ignore[assignment]
            right = new_node.right  # type: ignore[assignment]
            if (
                (
                    (left_f.prefix, right.prefix) == ("f", "")
                    or {left_f.prefix, right.prefix}.issubset({"r", "rf", "fr"})
                )
                and left_f.quote == right.quote
                and left_f.quote in {"'", '"'}
                and "{" not in right.raw_value
                and "}" not in right.raw_value
            ):
                return libcst.FormattedString(
                    lpar=new_node.lpar,
                    rpar=new_node.rpar,
                    parts=(
                        list(left_f.parts)
                        + [libcst.FormattedStringText(value=right.raw_value)]
                    ),
                    start=left_f.start,
                    end=left_f.end,
                )

        return new_node

    def leave_FormattedString(
        self,
        original_node: libcst.FormattedString,
        updated_node: libcst.FormattedString,
    ) -> libcst.FormattedString:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    # expressions

    def leave_Lambda(
        self, original_node: libcst.Lambda, updated_node: libcst.Lambda
    ) -> libcst.Lambda:
        # lambdas can have attributes:
        #   (lambda x: x + 1).__code__
        # and can be called (although it's sort of pointless):
        #   (lambda x: x + 1)(2)
        # and also may parse (for a human) less ambiguously with parens:
        #   (lambda x: x + 1), 2
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_Yield(
        self, original_node: libcst.Yield, updated_node: libcst.Yield
    ) -> libcst.Yield:
        new_node = updated_node
        # inject whitespace if a yield is used without trailing space
        # e.g. `yield('foo')` -> `yield ('foo')`
        if new_node.value is not None and (
            new_node.whitespace_after_yield.empty  # type: ignore[union-attr]
        ):
            new_node = new_node.with_changes(
                whitespace_after_yield=libcst.SimpleWhitespace(" ")
            )
        # expression nesting:
        #   return (yield from inner())
        if len(new_node.lpar) < 2:
            return new_node
        return self.modify_parenthesized_node(
            original_node,
            new_node,
            preserve_innermost=True,
        )

    def leave_GeneratorExp(
        self, original_node: libcst.GeneratorExp, updated_node: libcst.GeneratorExp
    ) -> libcst.GeneratorExp:
        # generator expression passed to a function call:
        #   foo((x % 2 for x in range(10)), bar)
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_IfExp(
        self, original_node: libcst.IfExp, updated_node: libcst.IfExp
    ) -> libcst.IfExp:
        # parens are required for usage of the results:
        #   (foo() if x else bar())["baz"]
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_Await(
        self, original_node: libcst.Await, updated_node: libcst.Await
    ) -> libcst.Await:
        # parens are required for usage of the results:
        #   (await foo())["bar"]
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_Match(
        self, original_node: libcst.Match, updated_node: libcst.Match
    ) -> libcst.Match:
        # inject whitespace if missing
        # ensures that `match(x): ...` converts to `match x: ...`
        if not original_node.whitespace_after_match.empty:
            return updated_node
        return updated_node.with_changes(
            whitespace_after_match=libcst.SimpleWhitespace(" ")
        )

    def leave_With(
        self, original_node: libcst.With, updated_node: libcst.With
    ) -> libcst.With:
        # remove parens from `with (foo): ...` as long as they are on the same line
        changes: dict[str, t.Any] = {}
        if original_node.whitespace_after_with.empty:
            changes["whitespace_after_with"] = libcst.SimpleWhitespace(" ")
        if self._singular_parens_are_same_line(original_node):
            changes["lpar"] = libcst.MaybeSentinel.DEFAULT
            changes["rpar"] = libcst.MaybeSentinel.DEFAULT
        if changes:
            return updated_node.with_changes(**changes)
        return updated_node

    def leave_If(self, original_node: libcst.If, updated_node: libcst.If) -> libcst.If:
        # inject whitespace if missing
        # ensures that `if(x): ...` converts to `if x: ...`
        if original_node.whitespace_before_test.empty:
            updated_node = updated_node.with_changes(
                whitespace_before_test=libcst.SimpleWhitespace(" ")
            )

        # convert any return of a variable after None checking to return None instead
        # i.e. `if x is None: return x` -> `if x is None: return None`
        if libcst.matchers.matches(
            original_node.test,
            libcst.matchers.Comparison(
                left=libcst.matchers.Name(),
                comparisons=[
                    libcst.matchers.ComparisonTarget(
                        operator=libcst.matchers.Is(),
                        comparator=libcst.matchers.Name("None"),
                    )
                ],
            ),
        ):
            var_name = original_node.test.left.value  # type: ignore[attr-defined]

            # if this is an indented block returning `$var_name`
            # swap it for an indented block returning `None`
            if libcst.matchers.matches(
                original_node.body,
                libcst.matchers.IndentedBlock(
                    body=[
                        libcst.matchers.SimpleStatementLine(
                            body=[
                                libcst.matchers.Return(
                                    value=libcst.matchers.Name(var_name),
                                )
                            ]
                        )
                    ]
                ),
            ):
                return_node = updated_node.body.body[0].body[0]  # type: ignore[union-attr]  # noqa: E501
                updated_return_node = return_node.with_changes(
                    value=libcst.Name("None")
                )
                updated_node = updated_node.with_changes(
                    body=updated_node.body.with_changes(
                        body=[
                            updated_node.body.body[0].with_changes(
                                body=[
                                    updated_return_node,
                                ]
                            )
                        ]
                    )
                )
            # if this is a statement (inline) returning `$var_name`
            # swap it for an statement returning `None`
            # also turn it into an indented block
            elif libcst.matchers.matches(
                original_node.body,
                libcst.matchers.SimpleStatementSuite(
                    body=[
                        libcst.matchers.Return(
                            value=libcst.matchers.Name(var_name),
                        )
                    ]
                ),
            ):
                trailing_comment = original_node.body.trailing_whitespace.comment  # type: ignore[attr-defined]  # noqa: E501
                updated_node = updated_node.with_changes(
                    body=libcst.IndentedBlock(
                        body=[
                            libcst.SimpleStatementLine(
                                body=[
                                    libcst.Return(
                                        value=libcst.Name("None"),
                                    )
                                ]
                            )
                        ],
                        header=(
                            libcst.TrailingWhitespace(
                                whitespace=(
                                    libcst.SimpleWhitespace(
                                        "  " if trailing_comment else ""
                                    )
                                ),
                                comment=trailing_comment,
                            )
                        ),
                    )
                )

        return updated_node

    def leave_FunctionDef(
        self, original_node: libcst.FunctionDef, updated_node: libcst.FunctionDef
    ) -> libcst.FunctionDef:
        if libcst.matchers.matches(
            original_node, MISSING_RETURN_ANNOTATION_INIT_MATCHER
        ):
            updated_node = updated_node.with_changes(
                returns=libcst.Annotation(annotation=libcst.Name("None"))
            )
        return updated_node

    def leave_ImportFrom(
        self, original_node: libcst.ImportFrom, updated_node: libcst.ImportFrom
    ) -> libcst.ImportFrom:
        changes: dict[str, t.Any] = {}
        if original_node.whitespace_after_import.empty:
            changes["whitespace_after_import"] = libcst.SimpleWhitespace(" ")
        if self._singular_parens_are_same_line(original_node):
            changes["lpar"] = None
            changes["rpar"] = None
        if changes:
            return updated_node.with_changes(**changes)
        return updated_node


def _make_paren_whitespace(
    last_line_spacing: str, *, comment: libcst.Comment | None = None
) -> libcst.ParenthesizedWhitespace:
    return libcst.ParenthesizedWhitespace(
        first_line=libcst.TrailingWhitespace(
            whitespace=libcst.SimpleWhitespace(value="  " if comment else ""),
            comment=comment,
            newline=libcst.Newline(),
        ),
        indent=True,
        last_line=libcst.SimpleWhitespace(value=last_line_spacing),
    )


def _convert_dict_element(arg: libcst.Arg) -> libcst.BaseDictElement:
    if arg.star == "":
        if not arg.keyword:
            raise ValueError(
                "Cannot convert dict arg which is non-splatted, non-keyword"
            )
        return libcst.DictElement(
            key=libcst.SimpleString(value=f'"{arg.keyword.value}"'),
            value=arg.value,
            comma=arg.comma,
        )
    return libcst.StarredDictElement(
        value=arg.value,
        comma=arg.comma,
    )


def _last_arg_whitespace(node: libcst.Call) -> libcst.BaseParenthesizableWhitespace:
    if not node.args:
        return libcst.SimpleWhitespace("")
    return node.args[-1].whitespace_after_arg


def _find_max_indent_of_concat_node(nodes: list[libcst.BaseString]) -> int:
    max_ = 0
    for node in nodes:
        if not isinstance(node, libcst.ConcatenatedString):
            continue
        ws_between = node.whitespace_between
        if not (hasattr(ws_between, "indent") and ws_between.indent):
            continue
        if not isinstance(ws_between, libcst.ParenthesizedWhitespace):
            continue
        max_ = max(max_, len(ws_between.last_line.value))
    return max_
